function [ K, z, y, mu, eta, relerr ] = em_isvm( xtrain, xtest, ytrain, lambda, s, c, nu, l, T, T0, maxK )
%input:
%   xtrain: real, n0 * d, training data
%   xtest: real, n1 * d, test data
%   ytrain: +-1, n0 * 1, training labels
%   lambda: real scalar, parameter; threshold for starting a new component
%   s: real scalar, parameter; trade-off between inference and learning
%   c: real scalar, parameter; regularization coefficient
%   nu: real scalar, parameter; variance of the prior distribution on
%   classifiers
%   l: loss, parameter
%   T: maximum # of iterations allowed
%   T0: minimum # of iteration required
%   maxK: maximum # of components allowed
%
%output:
%   K: # of nonempty components
%   z: {1,...,K}, n * 1, component assignments
%   y: +-1, n * 1, prediction
%   relerr: real scalar, relative error in loss function change

    % initialization
    n0 = size(xtrain, 1);
    n1 = size(xtest, 1);
    n = n0 + n1;
    d = size(xtrain, 2);
    x = [xtrain; xtest];
    ext_x = [x ones(n,1)];
    
    K = 1;
    z = ones(n, 1);
    iz = zeros(n, maxK);
    invomega = ones(n0, 1);
    mu = zeros(maxK, d);
    eta = zeros(maxK, d+1);
    
    mu(1, :) = mean(x);
    eta(1, :) = zeros(1, d+1);
    iz(:, 1) = ones(n, 1);
    zeta = zeros(n, 1);    
    
    pc_inf = zeros(n, maxK+1); %probability for components, inference
    pc_learn = zeros(n, maxK+1); %probability for components, learning
    priorL = 1/nu^2 * eye(d+1); 
    priorL(d+1, d+1) = 0;  %prior of eta variance 
    
    % compute eta^* and their corresponding thresholds
    for i = 1:n0
        eta_star = min(c*nu^2, 1/norm(x(i,:))^2) * ytrain(i) * ext_x(i, :); % 1 x d+1
        pc_learn(i, maxK+1) = norm(eta_star(1:d))^2/(2*nu^2) + 2*c*max(0, l - ytrain(i)*eta_star*ext_x(i,:)');
    end
    pc_inf(:, maxK+1) = lambda * ones(n, 1);
    
    fprintf('Initialization finished.\n');
    
    lastloss = 1;
    
    for t = 1:T
        
        % update eta
        for k = 1:K
            indicators = iz(1:n0, k); % n0 x 1
            weighted_indicators = iz(1:n0, k) .* invomega; % n0 x 1
            if (sum(indicators) == 0)
                eta(k, :) = zeros(1, d+1);
                continue;
            end
            alt_ext_x = ext_x(1:n0, :) .* weighted_indicators(:, ones(1, d+1)); % n0 x d+1
            Lambda = (priorL + c^2 * ext_x(1:n0, :)' * alt_ext_x)^(-1); % d+1 x d+1
            Mu = (c * (ytrain .* indicators)' * (ext_x(1:n0, :) + c*l*alt_ext_x)) * Lambda; 
            eta(k, :) = Mu;
        end 
        
        % compute margins and predict labels
        theta = sum((iz(:, 1:K) * eta(1:K, :)) .* ext_x, 2); % n x 1
        zeta = l*ones(n0, 1) - ytrain .* theta(1:n0); % n0 x 1
        y = (theta >= zeros(n, 1)) * 2 - 1; % n x 1, prediction
        
        % compute pc
        pc_inf(1:n, 1:K) = zeros(n, K);
        pc_learn(1:n, 1:K) = zeros(n, K);
        for k = 1:K
            % inference part
            mu0 = mu(k, :); %1 x d
            divmat = x - mu0(ones(1, n), :); % n by d, each row (x-mu0)
            divdist = sum(divmat .^ 2, 2); % n by 1
            pc_inf(:, k) = pc_inf(:, k) + s * divdist;
            % learning part
            ktheta = ext_x(1:n0, :) * eta(k, :)'; % n0 x 1
            kzeta = l*ones(n0, 1) - ytrain .* ktheta; % n0 x 1
            pc_learn(1:n0, k) = pc_learn(1:n0, k) + 2*c * max(zeros(n0, 1), kzeta);
        end
        
        % compute the loss function, and output it
        % term 1: regularization
        A = sum(sum(eta(1:K, 1:d) .^ 2, 2)) / (2*nu^2);
        % term 2 & 3: inference and learning
        B = 0;
        C = 0;
        for k = 1:K
            indicators = iz(1:n, k); % n x 1
            B = B + indicators' * pc_inf(:, k);
            C = C + indicators(1:n0)' * pc_learn(1:n0, k);
        end
        %term 4: AIC complexity
        D = lambda * K;
        
        relerr = abs(A+B+C+D - lastloss) / lastloss;
        lastloss = A+B+C+D;
        
        fprintf('Iter %d: K = %d, loss = %f, tr acc. = %f, inf/learn ratio = %f, relerr = %f\n', ...
            t, K, A+B+C+D, mean(y(1:n0) == ytrain), B/(A+C), relerr);
                        
        % update omega(inv)
        invomega = ones(n0, 1) ./ (c * abs(zeta)); % n0 x 1
        
        % update z
        if (K < maxK-1)
            pc_inf(:, K+1) = pc_inf(:, maxK+1);
            pc_learn(:, K+1) = pc_learn(:, maxK+1);
        else
            pc_inf(:, K+1) = 1e100 * ones(n, 1);
        end
        
        z0 = z;
        [minvalue, z] = min(pc_inf(:, 1:K+1) + pc_learn(:, 1:K+1), [], 2); % z: n x 1
        if ((mean(z == z0) == 1) && (t > T0))
            return;
        end
        
        K = max(z);
        for k = 1:K
            iz(:, k) = (z == k * ones(n, 1));
        end
        
        % update mu
        for k = 1:K
            indicators = iz(:, k); % n x 1
            if (sum(indicators) == 0)
                mu(k, :) = zeros(1, d);
            else
                mu(k, :) = sum(x .* indicators(:, ones(1, d))) / sum(indicators);
            end
        end
        
    end

end
